package dev.fyloz.trial.colorrecipesexplorer.service

import com.nhaarman.mockitokotlin2.*
import dev.fyloz.trial.colorrecipesexplorer.exception.EntityAlreadyExistsRestException
import dev.fyloz.trial.colorrecipesexplorer.exception.EntityNotFoundRestException
import dev.fyloz.trial.colorrecipesexplorer.model.*
import dev.fyloz.trial.colorrecipesexplorer.repository.MixTypeRepository
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import kotlin.test.assertEquals
import kotlin.test.assertFalse
import kotlin.test.assertTrue

class MixTypeServiceTest : AbstractNamedModelServiceTest<MixType, MixTypeService, MixTypeRepository>() {
    override val repository: MixTypeRepository = mock()
    private val materialService: MaterialService = mock()
    private val mixService: MixService = mock()
    override val service: MixTypeService = spy(MixTypeServiceImpl(repository, materialService, mixService))

    private val materialType: MaterialType = materialType()
    private val material: Material = material(id = 0L, materialType = materialType)
    override val entity: MixType = mixType(id = 0L, name = "mix type", material = material)
    override val anotherEntity: MixType = mixType(id = 1L, name = "another mix type")
    override val entityWithEntityName: MixType = mixType(id = 2L, name = entity.name)

    @AfterEach
    override fun afterEach() {
        reset(materialService)
        super.afterEach()
    }

    // existsByNameAndMaterialType

    @Test
    fun `existsByNameAndMaterialType() returns repository's answer`() {
        setOf(true, false).forEach {
            whenever(repository.existsByNameAndMaterialType(entity.name, materialType)).doReturn(it)

            val found = service.existsByNameAndMaterialType(entity.name, materialType)

            assertTrue { found == it }
        }
    }

    // getByMaterial()

    @Test
    fun `getByMaterial() returns the mix type with the given material`() {
        whenever(repository.findByMaterial(material)).doReturn(entity)

        val found = service.getByMaterial(material)

        assertEquals(entity, found)
    }

    @Test
    fun `getByMaterial() throws EntityNotFoundRestException when no mix type with the given material exists`() {
        whenever(repository.findByMaterial(material)).doReturn(null)

        val exception = assertThrows<EntityNotFoundRestException> { service.getByMaterial(material) }
        assertEquals(material.name, exception.value)
    }

    // getByNameAndMaterialType()

    @Test
    fun `getByNameAndMaterialType() returns the mix type with the given name and material type`() {
        whenever(repository.findByNameAndMaterialType(entity.name, materialType)).doReturn(entity)

        val found = service.getByNameAndMaterialType(entity.name, materialType)

        assertEquals(entity, found)
    }

    // getOrCreateForNameAndMaterialType()
    @Test
    fun `getOrCreateForNameAndMaterialType() calls getForNameAndMaterialType() when a mix type with the given name and material type exists`() {
        doReturn(true).whenever(service).existsByNameAndMaterialType(entity.name, materialType)
        doReturn(entity).whenever(service).getByNameAndMaterialType(entity.name, materialType)

        val found = service.getOrCreateForNameAndMaterialType(entity.name, materialType)

        verify(service).getByNameAndMaterialType(entity.name, materialType)

        assertEquals(entity, found)
    }

    @Test
    fun `getOrCreateForNameAndMaterialType() calls saveForNameAndMaterialType() when no mix type with the given name and material type exists`() {
        doReturn(false).whenever(service).existsByNameAndMaterialType(entity.name, materialType)
        doReturn(entity).whenever(service).saveForNameAndMaterialType(entity.name, materialType)

        val found = service.getOrCreateForNameAndMaterialType(entity.name, materialType)

        verify(service).saveForNameAndMaterialType(entity.name, materialType)

        assertEquals(entity, found)
    }

    // save()

    @Test
    fun `save() throws EntityAlreadyExistsRestException when a material with the name of the new mix type exists`() {
        whenever(materialService.existsByName(entity.name)).doReturn(true)

        val exception = assertThrows<EntityAlreadyExistsRestException> { service.save(entity) }
        assertEquals(entity.name, exception.value)
    }

    // saveForNameAndMaterialType()

    @Test
    fun `saveForNameAndMaterialType() creates a save a valid mix type with the given name and material type`() {
        val name = entity.name
        val materialType = materialType()

        doAnswer { it.arguments[0] }.whenever(service).save(any())

        val found = service.saveForNameAndMaterialType(name, materialType)

        verify(service).save(any())

        assertEquals(name, found.name)
        assertEquals(name, found.material.name)
        assertEquals(materialType, found.material.materialType)
        assertTrue(found.material.isMixType)
    }

    // updateForNameAndMaterialType()

    @Test
    fun `updateForNameAndMaterialType() updates the given mix type with the given name and material type`() {
        val mixType = mixType(id = 1L, material = material(isMixType = true))
        val name = entity.name
        val materialType = materialType()

        doAnswer { it.arguments[0] }.whenever(service).update(any())

        val found = service.updateForNameAndMaterialType(mixType, name, materialType)

        verify(service).update(any())

        assertEquals(mixType.id, found.id)
        assertEquals(name, found.name)
        assertEquals(name, found.material.name)
        assertEquals(materialType, found.material.materialType)
        assertTrue(found.material.isMixType)
    }

    // delete()

    override fun `deleteById() deletes the entity with the given id in the repository`() {
        whenever(repository.canBeDeleted(any())).doReturn(true)
        super.`deleteById() deletes the entity with the given id in the repository`()
    }
}
